use crate::common::args::Cli;

use std::{
    net::{SocketAddr, ToSocketAddrs},
    path::PathBuf,
    time::Duration,
};

use anyhow::{anyhow, Result};
use libpt::log::{error, trace};

const DEFAULT_TIMEOUT_LEN: u64 = 5000; // ms
const DEFAULT_DELAY_LEN: u64 = 200; // ms
const DEFAULT_WIN_AFTER: usize = 0xff;

#[derive(Clone)]
pub struct Config {
    pub addr: std::net::SocketAddr,
    pub hostname: String,
    pub timeout: Duration,
    pub delay: Duration,
    #[cfg(feature = "server")]
    pub win_after: usize,
    #[cfg(feature = "server")]
    pub key: Option<PathBuf>,
    pub certs: Option<PathBuf>,
    #[cfg(feature = "server")]
    pub indefinitely: bool,
}

impl Config {
    pub fn build(cli: &Cli) -> Result<Self> {
        let addr: SocketAddr = match cli.host.to_socket_addrs() {
            Ok(mut addr) => addr.next().unwrap(),
            Err(err) => {
                error!(
                    "could not resolve host {:?} to a socket address: {err:?}",
                    cli.host.clone()
                );
                return Err(anyhow!(
                    "could not resolve host {:?} to a socket address: {err:?}",
                    cli.host
                ));
            }
        };
        let hostname = match cli.host.split_once(':') {
            Some(hostname) => hostname.0.to_string(),
            None => return Err(anyhow!("malformatted host (no port specified)")),
        };
        trace!("config has resolved the given hostname to: {addr:?}");
        Ok(Config {
            addr,
            hostname,
            timeout: Duration::from_millis(DEFAULT_TIMEOUT_LEN),
            delay: Duration::from_millis(DEFAULT_DELAY_LEN),
            #[cfg(feature = "server")]
            win_after: DEFAULT_WIN_AFTER,
            #[cfg(feature = "server")]
            key: cli.key.clone(),
            certs: cli.certs.clone(),
            indefinitely: cli.indefinitely,
        })
    }
}
